import ast
import os
import queue
import re
import threading
import time
import traceback

from jupyter_client.manager import KernelManager

from .base import BaseLanguage

DEBUG_MODE = False
# Supresses a weird debugging error
os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"
# turn off colors in "terminal"
os.environ["ANSI_COLORS_DISABLED"] = "1"


class Python(BaseLanguage):
    file_extension = "py"
    name = "Python"

    def __init__(self):
        self.km = KernelManager()
        self.km.start_kernel()
        self.kc = self.km.client()
        self.kc.start_channels()
        while not self.kc.is_alive():
            time.sleep(0.1)
        time.sleep(0.5)

        self.listener_thread = None
        self.finish_flag = False

        # DISABLED because sometimes this bypasses sending it up to us for some reason!
        # Give it our same matplotlib backend
        # backend = matplotlib.get_backend()

        # Use Agg, which bubbles everything up as an image.
        # Not perfect (I want interactive!) but it works.
        backend = "Agg"

        code = f"""
        import matplotlib
        matplotlib.use('{backend}')
        """
        for _ in self.run(code):
            pass

        # DISABLED because it doesn't work??
        # Disable color outputs in the terminal, which don't look good in OI and aren't useful
        # code = """
        # from IPython.core.getipython import get_ipython
        # get_ipython().colors = 'NoColor'
        # """
        # self.run(code)

    def terminate(self):
        self.kc.stop_channels()
        self.km.shutdown_kernel()

    def run(self, code):
        # lel
        # exec(code)
        # return
        self.finish_flag = False
        try:
            try:
                preprocessed_code = self.preprocess_code(code)
            except Exception:
                # Any errors produced here are our fault.
                # Also, for python, you don't need them! It's just for active_line and stuff. Just looks pretty.
                preprocessed_code = code
            message_queue = queue.Queue()
            self._execute_code(preprocessed_code, message_queue)
            yield from self._capture_output(message_queue)
        except GeneratorExit:
            raise  # gotta pass this up!
        except Exception:
            content = traceback.format_exc()
            yield {"type": "console", "format": "output", "content": content}

    def _execute_code(self, code, message_queue):
        def iopub_message_listener():
            while True:
                # If self.finish_flag = True, and we didn't set it (we do below), we need to stop. That's our "stop"
                if self.finish_flag is True:
                    if DEBUG_MODE:
                        print("interrupting kernel!!!!!")
                    self.km.interrupt_kernel()
                    return
                try:
                    msg = self.kc.iopub_channel.get_msg(timeout=0.05)
                except queue.Empty:
                    continue

                if DEBUG_MODE:
                    print("-----------" * 10)
                    print("Message recieved:", msg["content"])
                    print("-----------" * 10)

                if msg["header"]["msg_type"] == "status" and msg["content"]["execution_state"] == "idle":
                    # Set finish_flag and return when the kernel becomes idle
                    if DEBUG_MODE:
                        print("from thread: kernel is idle")
                    self.finish_flag = True
                    return

                content = msg["content"]

                if msg["msg_type"] == "stream":
                    line, active_line = self.detect_active_line(content["text"])
                    if active_line:
                        message_queue.put(
                            {
                                "type": "console",
                                "format": "active_line",
                                "content": active_line,
                            }
                        )
                    message_queue.put({"type": "console", "format": "output", "content": line})
                elif msg["msg_type"] == "error":
                    content = "\n".join(content["traceback"])
                    # Remove color codes
                    ansi_escape = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
                    content = ansi_escape.sub("", content)
                    message_queue.put(
                        {
                            "type": "console",
                            "format": "output",
                            "content": content,
                        }
                    )
                elif msg["msg_type"] in ["display_data", "execute_result"]:
                    data = content["data"]
                    if "image/png" in data:
                        message_queue.put(
                            {
                                "type": "image",
                                "format": "base64.png",
                                "content": data["image/png"],
                            }
                        )
                    elif "image/jpeg" in data:
                        message_queue.put(
                            {
                                "type": "image",
                                "format": "base64.jpeg",
                                "content": data["image/jpeg"],
                            }
                        )
                    elif "text/html" in data:
                        message_queue.put(
                            {
                                "type": "code",
                                "format": "html",
                                "content": data["text/html"],
                            }
                        )
                    elif "text/plain" in data:
                        message_queue.put(
                            {
                                "type": "console",
                                "format": "output",
                                "content": data["text/plain"],
                            }
                        )
                    elif "application/javascript" in data:
                        message_queue.put(
                            {
                                "type": "code",
                                "format": "javascript",
                                "content": data["application/javascript"],
                            }
                        )

        self.listener_thread = threading.Thread(target=iopub_message_listener)
        # self.listener_thread.daemon = True
        self.listener_thread.start()

        if DEBUG_MODE:
            print("thread is on:", self.listener_thread.is_alive(), self.listener_thread)

        self.kc.execute(code)

    def detect_active_line(self, line):
        if "##active_line" in line:
            # Split the line by "##active_line" and grab the last element
            last_active_line = line.split("##active_line")[-1]
            # Split the last active line by "##" and grab the first element
            active_line = int(last_active_line.split("##")[0])
            # Remove all ##active_line{number}##\n
            line = re.sub(r"##active_line\d+##\n", "", line)
            return line, active_line
        return line, None

    def _capture_output(self, message_queue):
        while True:
            if self.listener_thread:
                try:
                    output = message_queue.get(timeout=0.1)
                    if DEBUG_MODE:
                        print(output)
                    yield output
                except queue.Empty:
                    if self.finish_flag:
                        if DEBUG_MODE:
                            print("we're done")
                        break
            time.sleep(0.1)

    def stop(self):
        self.finish_flag = True

    def preprocess_code(self, code):
        return preprocess_python(code)


def preprocess_python(code):
    """
    Add active line markers
    Wrap in a try except
    """

    code = code.strip()

    # Add print commands that tell us what the active line is
    # but don't do this if any line starts with ! or %
    if not any(line.strip().startswith(("!", "%")) for line in code.split("\n")):
        code = add_active_line_prints(code)

    # Wrap in a try except (DISABLED)
    # code = wrap_in_try_except(code)

    # Remove any whitespace lines, as this will break indented blocks
    # (are we sure about this? test this)
    code_lines = code.split("\n")
    code_lines = [c for c in code_lines if c.strip() != ""]
    code = "\n".join(code_lines)

    return code


def add_active_line_prints(code):
    """
    Add print statements indicating line numbers to a python string.
    """
    # Replace newlines and comments with pass statements, so the line numbers are accurate (ast will remove them otherwise)
    code_lines = code.split("\n")
    in_multiline_string = False
    for i in range(len(code_lines)):
        line = code_lines[i]
        if '"""' in line or "'''" in line:
            in_multiline_string = not in_multiline_string
        if not in_multiline_string and (line.strip().startswith("#") or line == ""):
            whitespace = len(line) - len(line.lstrip(" "))
            code_lines[i] = " " * whitespace + "pass"
    processed_code = "\n".join(code_lines)
    try:
        tree = ast.parse(processed_code)
    except Exception:
        # If you can't parse the processed version, try the unprocessed version before giving up
        tree = ast.parse(code)
    transformer = AddLinePrints()
    new_tree = transformer.visit(tree)
    return ast.unparse(new_tree)


class AddLinePrints(ast.NodeTransformer):
    """
    Transformer to insert print statements indicating the line number
    before every executable line in the AST.
    """

    def insert_print_statement(self, line_number):
        """Inserts a print statement for a given line number."""
        return ast.Expr(
            value=ast.Call(
                func=ast.Name(id="print", ctx=ast.Load()),
                args=[ast.Constant(value=f"##active_line{line_number}##")],
                keywords=[],
            )
        )

    def process_body(self, body):
        """Processes a block of statements, adding print calls."""
        new_body = []

        # In case it's not iterable:
        if not isinstance(body, list):
            body = [body]

        for sub_node in body:
            if hasattr(sub_node, "lineno"):
                new_body.append(self.insert_print_statement(sub_node.lineno))
            new_body.append(sub_node)

        return new_body

    def visit(self, node):
        """Overridden visit to transform nodes."""
        new_node = super().visit(node)

        # If node has a body, process it
        if hasattr(new_node, "body"):
            new_node.body = self.process_body(new_node.body)

        # If node has an orelse block (like in for, while, if), process it
        if hasattr(new_node, "orelse") and new_node.orelse:
            new_node.orelse = self.process_body(new_node.orelse)

        # Special case for Try nodes as they have multiple blocks
        if isinstance(new_node, ast.Try):
            for handler in new_node.handlers:
                handler.body = self.process_body(handler.body)
            if new_node.finalbody:
                new_node.finalbody = self.process_body(new_node.finalbody)

        return new_node


def wrap_in_try_except(code):
    # Add import traceback
    code = "import traceback\n" + code

    # Parse the input code into an AST
    parsed_code = ast.parse(code)

    # Wrap the entire code's AST in a single try-except block
    try_except = ast.Try(
        body=parsed_code.body,
        handlers=[
            ast.ExceptHandler(
                type=ast.Name(id="Exception", ctx=ast.Load()),
                name=None,
                body=[
                    ast.Expr(
                        value=ast.Call(
                            func=ast.Attribute(
                                value=ast.Name(id="traceback", ctx=ast.Load()),
                                attr="print_exc",
                                ctx=ast.Load(),
                            ),
                            args=[],
                            keywords=[],
                        )
                    ),
                ],
            )
        ],
        orelse=[],
        finalbody=[],
    )

    # Assign the try-except block as the new body
    parsed_code.body = [try_except]

    # Convert the modified AST back to source code
    return ast.unparse(parsed_code)



#!/usr/bin/env python3

import concurrent.futures
import heapq
import itertools
import math
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Tuple

from loguru import logger

from uglychain.storage import DillStorage, Storage
from uglychain.utils import segment

from .base import StorageRetriever


class PathNotFoundError(Exception):
    pass


@dataclass
class BM25:
    k1: float = 1.5
    b: float = 0.75
    preprocessed_texts: List[str] = field(default_factory=list)
    word_sets: List[Set[str]] = field(default_factory=list)
    text_lens: List[int] = field(default_factory=list)
    tf_values: dict = field(default_factory=dict)
    idf_values: dict = field(default_factory=dict)
    sum_len: float = field(default=0)

    def calculate_tf(self, word: str, text: str) -> float:
        return text.split().count(word) / len(text.split())

    def calculate_idf(self, word: str) -> float:
        matches = len([True for text in self.preprocessed_texts if word in text.split()])
        return math.log(len(self.preprocessed_texts) / matches) if matches else 0.0

    def calculate_bm25_score(self, i: int, query: str) -> float:
        score = 0
        preprocessed_query = segment(query).split()
        for word in preprocessed_query:
            if word not in self.word_sets[i]:
                continue
            key = f"{word}_{i}"
            tf_value = self.tf_values.get(key, 1e-9)
            idf_value = self.idf_values.get(word, 0)
            tf_idf_value = tf_value * idf_value
            text_len = self.text_lens[i]
            avg_len = self.sum_len / len(self.preprocessed_texts) if self.preprocessed_texts else 1
            score_part = (self.k1 + 1) / (tf_value + self.k1 * (1 - self.b + self.b * text_len / avg_len))
            score += tf_idf_value * score_part
        return score

    def search(self, query: str, n: int = StorageRetriever.default_n) -> List[Tuple[int, float]]:
        num = len(self.text_lens)
        with concurrent.futures.ThreadPoolExecutor() as executor:
            scores = list(executor.map(self.calculate_bm25_score, range(num), itertools.repeat(query)))
        scores = list(zip(range(num), scores, strict=False))
        top_n_scores = heapq.nlargest(n, scores, key=lambda x: x[1])
        return top_n_scores

    def add(self, text: str) -> None:
        text_id = len(self.text_lens)
        preprocessed_text = segment(text)
        preprocessed_text_split = preprocessed_text.split()
        self.preprocessed_texts.append(preprocessed_text)
        self.sum_len += len(preprocessed_text_split)
        self.word_sets.append(set(preprocessed_text_split))
        self.text_lens.append(len(preprocessed_text_split))
        for word in self.word_sets[text_id]:
            key = f"{word}_{text_id}"
            self.tf_values[key] = self.calculate_tf(word, preprocessed_text)
            if word not in self.idf_values:
                self.idf_values[word] = self.calculate_idf(word)


@dataclass
class BM25Retriever(StorageRetriever):
    storage: Storage = field(default_factory=DillStorage)
    texts: List[str] = field(init=False, default_factory=list)
    metadatas: List[Dict[str, str]] = field(init=False, default_factory=list)
    _data: BM25 = field(init=False)

    def search(self, query: str, n: int = StorageRetriever.default_n) -> List[str]:
        if not query or self.is_empty:
            return []
        top_n_scores = self._data.search(query, n)
        return [self.texts[i] for i, _ in top_n_scores]

    def add(self, text: str, metadata: Optional[Dict[str, str]] = None) -> None:
        if not text:
            logger.warning("Text cannot be empty.")
            return
        if text in self.texts:
            logger.warning(f"Text already exists: {text}")
            return
        self.texts.append(text)
        self.metadatas.append(metadata or {})
        self._data.add(text)
        self._save()

    def init(self) -> None:
        self._data = BM25()
        self.texts = []
        self.metadatas = []
        self._save()

    def _save(self) -> None:
        data = self._data.__dict__.copy()
        data["texts"] = self.texts
        data["metadatas"] = self.metadatas
        self.storage.save(data)

    def _load(self) -> None:
        try:
            data = self.storage.load()
            self.texts = data.pop("texts")
            self.metadatas = data.pop("metadatas")
            self._data = BM25(**data)
            return
        except Exception:
            self.init()

    @property
    def is_empty(self):
        return not self.texts

    def all(self) -> List[str]:
        return self.texts



import sqlite3
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List

from .base import Storage


@dataclass
class SQLiteStorage(Storage):
    file: str = "data/cache.db"
    table: str = "cache"
    expirationIntervalInDays: int = 10

    def __post_init__(self):
        path = Path(self.file)
        with sqlite3.connect(path) as conn:
            cur = conn.cursor()
            cur.execute(
                f"CREATE TABLE IF NOT EXISTS {self.table} (key TEXT PRIMARY KEY, value TEXT, timestamp TEXT Not NULL DEFAULT (date('now','localtime')))"
            )
        self._conn = sqlite3.connect(path)
        self._cur = self._conn.cursor()

    def save(self, data: Dict[str, str]):
        self._cur.executemany(
            f"""
            INSERT OR REPLACE INTO {self.table} (key, value)
            VALUES (?, ?)
        """,
            data.items(),
        )
        self._conn.commit()

    def load(self, keys: List[str] | str | None, condition: str | None = None) -> Dict[str, str]:
        if keys is None:
            query_sql = (
                f"SELECT key, value FROM {self.table} WHERE date('now', 'localtime') < date(timestamp, '+? day')"
            )
            if condition is not None:
                query_sql += f" and {condition}"
            self._cur.execute(query_sql, (self.expirationIntervalInDays,))
            return {row[0]: row[1] for row in self._cur.fetchall()}
        if isinstance(keys, str):
            keys = [keys]

        placeholders = ", ".join("?" for _ in keys)
        params = keys + [str(self.expirationIntervalInDays)]
        query_sql = f"SELECT key, value FROM {self.table} WHERE key IN ({placeholders}) and date('now', 'localtime') < date(timestamp, '+' || ? || ' day')"
        if condition is not None:
            query_sql += f" and {condition}"
        self._cur.execute(query_sql, params)

        return {row[0]: row[1] for row in self._cur.fetchall()}



from dataclasses import dataclass, field
from typing import Any, Dict

from .base import Storage


@dataclass
class LlamaIndexStorage(Storage):
    persist_dir: str = "data/llama-index"
    kwargs: Dict = field(default_factory=dict)

    def save(self, index: Any):
        if self.kwargs:
            index.storage_context.persist(**self.kwargs)
        else:
            index.storage_context.persist(persist_dir=self.persist_dir)

    def load(self) -> Any:
        try:
            from llama_index.core import (
                StorageContext,
                load_index_from_storage,
            )
        except ImportError as err:
            raise ImportError("You need to install `pip install llama-index-core` to use this storage.") from err
        if self.kwargs:
            storage_context = StorageContext.from_defaults(**self.kwargs)
        else:
            storage_context = StorageContext.from_defaults(persist_dir=self.persist_dir)
        index = load_index_from_storage(storage_context)
        return index



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, cast

from uglychain.storage import LlamaIndexStorage
from uglychain.utils import config

from .base import BaseRetriever, StorageRetriever

try:
    from llama_index.core import Settings
    from llama_index.embeddings.openai import OpenAIEmbedding

    Settings.embed_model = OpenAIEmbedding(
        api_base=config.embedding_api_base, api_key=config.embedding_api_key, model=config.embedding_model
    )
except Exception:
    pass


@dataclass
class LlamaIndexRetriever(BaseRetriever):
    index: Any = None
    query_kwargs: Dict = field(default_factory=dict)

    def __post_init__(self):
        assert self.index is not None, "You need to set `index` to use this retriever."
        try:
            from llama_index.core.indices.base import BaseGPTIndex
        except ImportError as err:
            raise ImportError(
                "You need to install `pip install llama-index-core llama-index-embeddings-openai` to use this retriever."
            ) from err
        self.index = cast(BaseGPTIndex, self.index)

    def search(self, query: str, n: int = BaseRetriever.default_n) -> List[str]:
        retriever = self.index.as_retriever(similarity_top_k=n, **self.query_kwargs)
        response = retriever.retrieve(query)

        # parse source nodes
        docs = []
        for source_node in response:
            docs.append(source_node.text)
        return docs


@dataclass
class LlamaIndexStorageRetriever(LlamaIndexRetriever, StorageRetriever):
    persist_dir: str = ""
    storage: LlamaIndexStorage = field(default_factory=LlamaIndexStorage)

    def __post_init__(self):
        if self.persist_dir:
            self.storage = LlamaIndexStorage(persist_dir=self.persist_dir)
        StorageRetriever.__post_init__(self)

    def init(self):
        try:
            from llama_index.core import VectorStoreIndex
        except ImportError as err:
            raise ImportError("You need to install `pip install llama-index-core` to use this retriever.") from err
        self.index = VectorStoreIndex.from_documents([])
        self.storage.save(self.index)

    def add(self, text: str, metadata: Optional[Dict[str, str]] = None):
        if not metadata:
            metadata = {}
        try:
            from llama_index.core import Document
        except ImportError as err:
            raise ImportError("You need to install `pip install llama-index-core` to use this retriever.") from err
        self.index.insert(Document(text=text, extra_info=metadata))
        self.storage.save(self.index)

    def _load(self):
        try:
            self.index = self.storage.load()
        except Exception:
            self.init()



from typing import Literal

from .languages.applescript import AppleScript
from .languages.javascript import JavaScript
from .languages.powershell import PowerShell
from .languages.r import R
from .languages.shell import Shell

try:
    from .languages.ipython import Python
except ImportError:
    from .languages.python import Python

LANGUAGES = [
    Shell,
    JavaScript,
    AppleScript,
    R,
    PowerShell,
    Python,
]


def get_language(language: str):
    for lang in LANGUAGES:
        if language.lower() == lang.name.lower() or (hasattr(lang, "aliases") and language in lang.aliases):
            return lang
    return None


def run_code(code: str, language: Literal["Shell", "Powershell", "Applescript", "Python", "R", "Javascript"]) -> str:
    """This function allows you to execute code **on the user's machine** and retrieve the terminal output. Notice the Python code is sent to a Jupyter kernel for execution.

    Args:
        language (str): The programming language
        code (str): The code to execute
    """
    active_languages = {}
    if language not in active_languages:
        active_languages[language] = get_language(language)()  # type: ignore

    output_messages = []
    for chunk in active_languages[language].run(code):
        if chunk.get("format") != "active_line":
            if (
                output_messages != []
                and output_messages[-1].get("type") == chunk["type"]
                and output_messages[-1].get("format") == chunk["format"]
            ):
                output_messages[-1]["content"] += chunk["content"]
            else:
                output_messages.append(chunk)

    return "\n".join(a["content"] for a in output_messages)



#!/usr/bin/env python3

from dataclasses import dataclass
from typing import List
from xml.etree import ElementTree

import requests
from loguru import logger
from tenacity import RetryError, before_sleep_log, retry, stop_after_attempt, wait_fixed

from .base import BaseRetriever

ARXIV_SEARCH_API_URL = "http://export.arxiv.org/api/query"


@dataclass
class ArxivRetriever(BaseRetriever):
    def search(self, query: str, n: int = BaseRetriever.default_n) -> List[str]:
        try:
            response = self._send_request(query, n)
            xml = response.text
            root = ElementTree.fromstring(xml)
            entries = root.findall("{http://www.w3.org/2005/Atom}entry")
            results = [self._format(entry) for entry in entries[:n]]
            return results
        except ElementTree.ParseError as e:
            logger.error(f"Error occurred while parsing XML: {e}")
            return []
        except (requests.RequestException, RetryError) as e:
            logger.error(f"Error occurred while sending request or parsing response: {e}")
            return []

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_fixed(1),
        before_sleep=before_sleep_log(logger, "WARNING"),  # type: ignore
    )
    def _send_request(self, query: str, n: int):
        response = requests.get(
            ARXIV_SEARCH_API_URL,
            params={"search_query": query, "start": 0, "max_results": n},
            timeout=5,
        )
        response.raise_for_status()
        return response

    @classmethod
    def _format(cls, entry: ElementTree.Element) -> str:
        title = entry.find("{http://www.w3.org/2005/Atom}title").text  # type: ignore
        summary = entry.find("{http://www.w3.org/2005/Atom}summary").text  # type: ignore
        link = entry.find('{http://www.w3.org/2005/Atom}link[@title="pdf"]').get("href")  # type: ignore
        if not all([title, link, summary]):
            logger.error(f"Missing key in search result: {entry}")
            raise KeyError("Missing key in search result")
        return f"[{title}]({link})\n> {summary}"



import inspect
from dataclasses import dataclass, field
from typing import Callable, List, Optional, Union, cast

from docstring_parser import parse
from pydantic import BaseModel, Field

from uglychain import LLM
from uglychain.tools import text_completion

from .base import BaseWorker

ROLE = "You are an expert task manager AI."

CREATE_TASKS_PROMPT = """You are an expert task list creation AI tasked with creating a list of tasks, considering the ultimate objective of your team: {objective}.
Create a very short task list based on the objective, the final output of the last task will be provided back to the user. Limit tasks types to those that can be completed with the available tools listed below. Task description should be detailed.

### AVAILABLE TOOLS: {tools_info}.

### RULES:
- The final task should always output the final result of the overall objective.
- Make sure all task IDs are in chronological order.
- Helpful Notes as guidance: "{notes}"
- no more than 5 tasks should be created.
- task descriptions should be in Chinese.
"""

TASK_MANAGER_PROMPT = """
You are a task management AI tasked with reprioritizing the a list of tasks, considering the ultimate objective of your team: {objective}. Create new tasks based on the result of last task if necessary for the objective. Limit tasks types to those that can be completed with the available tools listed below. Task description should be detailed.

### AVAILABLE TOOLS: {tools_info}.

### RULES:
- Do not remove any tasks.
- The maximum task list length is 7. Do not add an 8th task.
- Do not reorder completed tasks. Only reorder and dedupe incomplete tasks.
- Make sure all task IDs are in chronological order.
- The last step is always to provide a final summary report of all tasks.

### Here is the current task list: {tasks}
### The last completed task has the following result: {task_output}.
"""


def get_function_description(func: Callable) -> str:
    try:
        return parse(inspect.getdoc(func)).short_description  # type: ignore
    except AttributeError:
        return ""


class Task(BaseModel):
    id: int = Field(..., description="The task id.")
    task: str = Field(..., description="The task description.")
    dependent_task_ids: List[int] = Field(
        default_factory=list,
        description="The task ids that this task is dependent on. It should always be an empty array, or an array of int representing the task ID it should pull results from.",
    )
    completed: bool = Field(
        default=False,
        description="The task completion status.",
    )

    @property
    def result(self) -> str:
        if not hasattr(self, "_result"):
            self._result = ""
        return self._result

    @result.setter
    def result(self, value: str):
        self._result = value


class Tasks(BaseModel):
    tasks: List[Task] = Field(
        ...,
        description="The list of tasks.",
    )

    def get_task_by_id(self, task_id: int) -> Task:
        try:
            return next(task for task in self.tasks if task.id == task_id)
        except StopIteration as err:
            raise ValueError(f"Task with id {task_id} not found.") from err

    def execute_task(self, task: Union[Task, int], execute: Callable[[str, str, str], str], objective: str):
        if isinstance(task, int):
            task = self.get_task_by_id(task)
        assert not task.completed, f"Task {task.id} is already completed."
        dependent_tasks = [self.get_task_by_id(dependent_task_id) for dependent_task_id in task.dependent_task_ids]
        assert all(
            dependent_task.completed for dependent_task in dependent_tasks
        ), f"Dependent tasks for task {task.id} are not completed."
        dependent_task_outputs = "\n".join(dependent_task.result for dependent_task in dependent_tasks)
        result = execute(task.task, dependent_task_outputs, objective)
        task.completed = True
        task.result = result
        return result

    def reorder_tasks(self):
        self.tasks.sort(key=lambda x: x.id)

    def __str__(self):
        tasks = []
        for task in self.tasks:
            task_dict = task.model_dump()
            task_dict["completed"] = task.completed
            tasks.append(task_dict)
        return str(tasks)

    @property
    def pretty(self):
        output = "\n"
        for task in self.tasks:
            output += f"{'✓' if task.completed else '✗'} {task.id}. {task.task} [ {','.join(str(i) for i in task.dependent_task_ids)} ]\n"
        return output


@dataclass
class Planner(BaseWorker):
    role: Optional[str] = ROLE
    prompt: str = field(init=False)
    tools: List[Callable] = field(default_factory=list)
    objective: str = ""
    llm: LLM = field(init=False)

    def __post_init__(self):
        self.llm = LLM(CREATE_TASKS_PROMPT, self.model, self.role, Tasks)
        try:
            self.tools.index(text_completion)
        except ValueError:
            self.tools.insert(0, text_completion)
        self.tools_info = ", ".join([f"`{tool.__name__}`:'{get_function_description(tool)}'" for tool in self.tools])

    def run(
        self,
        objective: str = "",
        tasks: Optional[Union[Tasks, List[str]]] = None,
        task_output: str = "",
    ):
        if tasks is None or isinstance(tasks, list):
            assert not task_output, "task_output should be empty if tasks is a list."
            self.llm.prompt = CREATE_TASKS_PROMPT
            if tasks:
                notes = ";".join(f"{i+1}. {item}" for i, item in enumerate(tasks))
            else:
                notes = ""
            response = self._ask(objective=objective, tools_info=self.tools_info, notes=notes)
            return cast(Tasks, response)
        else:
            self.llm.prompt = TASK_MANAGER_PROMPT
            response = self._ask(objective=objective, tools_info=self.tools_info, task_output=task_output, tasks=tasks)
            return cast(Tasks, response)



from dataclasses import dataclass
from typing import Any, Dict

from uglychain.utils import config

from .openai_api import ChatGPTAPI


@dataclass
class Baichuan(ChatGPTAPI):
    api_key: str = config.baichuan_api_key
    base_url: str = "https://api.baichuan-ai.com/v1"
    name: str = "Baichuan"
    use_max_tokens: bool = False

    @property
    def default_params(self) -> Dict[str, Any]:
        kwargs = super().default_params
        kwargs.pop("frequency_penalty")
        return kwargs



#!/usr/bin/env python3

from .base import BaseLanguageModel
from .instructor import Instructor, ParseError
from .tools import FunctionCall, finish, run_function

__all__ = [
    "BaseLanguageModel",
    "Instructor",
    "ParseError",
    "FunctionCall",
    "finish",
    "run_function",
]



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Type, Union

from loguru import logger
from pydantic import BaseModel

from uglychain.llm import BaseLanguageModel
from uglychain.utils import config, retry_decorator


@dataclass
class Ollama(BaseLanguageModel):
    top_p: float = 0.9

    def generate(
        self,
        prompt: str = "",
        response_model: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Callable]] = None,
        stop: Union[Optional[str], List[str]] = None,
    ) -> str:
        kwargs = self.get_kwargs(prompt, response_model, tools, stop)
        response = self.completion_with_backoff(**kwargs)
        logger.trace(f"kwargs:{kwargs}\nresponse:{response['message']}")
        return response["message"]["content"]

    def get_kwargs(
        self,
        prompt: str,
        response_model: Optional[Type[BaseModel]],
        tools: Optional[List[Callable]],
        stop: Union[Optional[str], List[str]],
    ) -> Dict[str, Any]:
        kwargs = super().get_kwargs(prompt, response_model, tools, stop)
        stop_value = kwargs.pop("stop")
        kwargs["options"]["stop"] = stop_value
        if response_model:
            kwargs["format"] = "json"
        return kwargs

    @property
    def default_params(self) -> Dict[str, Any]:
        options = {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
            "seed": self.seed,
        }
        kwargs = {
            "model": self.model,
            "options": options,
        }
        return kwargs

    def _create_client(self):
        try:
            from ollama import Client
        except ImportError as err:
            raise ImportError("You need to install `pip install ollama` to use this provider.") from err
        return Client(host=config.ollama_host)

    @retry_decorator()
    def completion_with_backoff(self, **kwargs):
        return self.client.chat(**kwargs)

    @property
    def max_tokens(self):
        return 2000



#!/usr/bin/env python3

from dataclasses import dataclass
from typing import List

from loguru import logger
from pydantic import BaseModel

from uglychain.chains import LLM
from uglychain.provider import Model

from .base import BaseRetriever
from .model import Retriever

PROMPT = """A question is provided below. Given the question, extract up to {max_keywords} keywords from the text. Focus on extracting the keywords that we can use to best lookup answers to the question. Avoid stopwords. If the keywords is not in English, please translate it to English.
---------------------
{query}
---------------------
"""


class Keywords(BaseModel):
    keywords: List[str]


@dataclass
class CombineRetriever(BaseRetriever):
    retrievers: list[Retriever]
    model: Model = Model.DEFAULT

    def search(self, query: str, n: int = BaseRetriever.default_n) -> List[str]:
        results = []
        if Retriever.Arxiv in self.retrievers or Retriever.Bing in self.retrievers:
            keywords = self.get_kewords(query)
            query_keywords = " ".join(keywords)
            logger.trace(f"query_keywords: {query_keywords}")
        for retriever in self.retrievers:
            instance = retriever()
            if retriever in {Retriever.Arxiv, Retriever.Bing}:
                results.extend(instance.search(query_keywords, n))
            else:
                results.extend(instance.search(query, n))
        return results

    def get_kewords(self, query: str, n: int = 5) -> List[str]:
        llm = LLM(PROMPT, self.model, response_model=Keywords)
        response = llm(query=query, max_keywords=n)
        return response.keywords[:5]



from dataclasses import dataclass
from pathlib import Path
from typing import Any

import dill

from .base import Storage


@dataclass
class DillStorage(Storage):
    file: str = "data/temp.pkl"

    def __post_init__(self):
        self.path = Path(self.file)

    def save(self, data: Any):
        with open(self.path, "wb") as f:
            dill.dump(data, f)

    def load(self) -> Any:
        with open(self.path, "rb") as f:
            return dill.load(f)



#!/usr/bin/env python3

from dataclasses import dataclass
from typing import Callable, List

from .base import BaseRetriever


@dataclass
class CustomRetriever(BaseRetriever):
    search_function: Callable[[str, int], List[str]]

    def search(self, query: str, n: int = BaseRetriever.default_n) -> List[str]:
        return self.search_function(query, n)



#!/usr/bin/env python3

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Optional

from uglychain import LLM, Model
from uglychain.retrievers import BaseRetriever
from uglychain.storage import Storage

RETRIEVER_PROMPT = """# Context from Retriever(Maybe useful for you to resolve the problem):
---
{retriever_context}
---
"""


@dataclass
class BaseWorker(ABC):
    """Base class for actions.

    Attributes:
        role: The role of the action.
        filename: The filename associated with the action.
        llm: The LLMChain object used for the action.
    """

    role: Optional[str] = None
    prompt: str = "{prompt}"
    model: Model = Model.DEFAULT
    retriever: Optional[BaseRetriever] = None
    storage: Optional[Storage] = None
    llm: Optional[LLM] = field(default=None, init=False)

    def __post_init__(self):
        if self.retriever:
            self.prompt = RETRIEVER_PROMPT + self.prompt

    def _ask(self, *args, **kwargs) -> Any:
        """Ask a question to the LLMChain object.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            The response from the LLMChain object.
        """
        assert self.llm, "llm is required"
        response = self.llm(*args, **kwargs)
        return response

    @abstractmethod
    def run(self, *args, **kwargs):
        """Abstract method to be implemented by subclasses.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.
        """
        if not self.llm:
            self.llm = LLM(self.prompt, self.model, self.role)
        response = self._ask(*args, **kwargs)
        if self.storage:
            self.storage.save(response)
        return response



stop_words = {
    "着呢",
    "doing",
    "大批",
    "与",
    "serious",
    "likely",
    "every",
    "“",
    "right",
    "嘎登",
    "如果",
    "am",
    "非但",
    "specified",
    "就是",
    "进入",
    "might",
    "尔后",
    "appear",
    "先後",
    "关于",
    "definitely",
    "根据",
    "me",
    "my",
    "et",
    "clearly",
    "她的",
    "needs",
    "每天",
    "soon",
    "哟",
    "不怕",
    "little",
    "中小",
    "如何",
    "必要",
    "最近",
    "照着",
    "充分",
    "useful",
    "seem",
    "呼哧",
    "respectively",
    "myself",
    "属于",
    "宣布",
    "gives",
    "广大",
    "is",
    "thence",
    "你的",
    "therefore",
    "exactly",
    "contain",
    "edu",
    "果然",
    "扩大",
    "嗡嗡",
    "个人",
    "咳",
    "against",
    "反过来",
    "most",
    "哪",
    "一样",
    "almost",
    "後面",
    "规定",
    "不比",
    "cant",
    "不仅",
    "none",
    "besides",
    "由于",
    "known",
    "当前",
    "toward",
    "随着",
    "inner",
    "乎",
    "迅速",
    "你们",
    "您",
    "won't",
    "这会儿",
    "换句话说",
    "outside",
    "went",
    "防止",
    "进步",
    "towards",
    "except",
    "现代",
    "each",
    "a's",
    "are",
    "反过来说",
    "may",
    "不过",
    "让",
    "前面",
    "hello",
    "theirs",
    "表示",
    "有时",
    "首先",
    "大大",
    "some",
    "help",
    "下去",
    "不敢",
    "各",
    "特点",
    "先生",
    "hereupon",
    "甚么",
    "there's",
    "大量",
    "their",
    "value",
    "for",
    "not",
    "ours",
    "宁可",
    "几乎",
    "us",
    "latter",
    "每当",
    "罢了",
    "曾经",
    "thats",
    "省得",
    "这些",
    "别的",
    "换言之",
    "要求",
    "良好",
    "嘛",
    "如",
    "meanwhile",
    "这点",
    "纵使",
    "particularly",
    "大家",
    "actually",
    "集中",
    "different",
    "继续",
    "several",
    "do",
    "它的",
    "because",
    "还是",
    "自",
    "及",
    "next",
    "else",
    "whole",
    "嘘",
    "显然",
    "believe",
    "吧",
    "失去",
    "产生",
    "归",
    "加入",
    "as",
    "各级",
    "必须",
    "似的",
    "另一方面",
    "给",
    "慢说",
    "i'm",
    "okay",
    "anyhow",
    "丰富",
    "岂但",
    "nobody",
    "随",
    "immediate",
    "果真",
    "哪里",
    "与此同时",
    "直到",
    "sorry",
    "了解",
    "himself",
    "sup",
    "with",
    "各人",
    "以后",
    "indicate",
    "wasn't",
    "than",
    "以至于",
    "consider",
    "当然",
    "复杂",
    "hardly",
    "一",
    "多数",
    "即",
    "c's",
    "而言",
    "asking",
    "应当",
    "thoroughly",
    "via",
    "apart",
    "依",
    "gotten",
    "那麽",
    "happens",
    "起见",
    "假若",
    "哪怕",
    "now",
    "漫说",
    "比如",
    "论",
    "一天",
    "somehow",
    "哪个",
    "inasmuch",
    "假使",
    "amongst",
    "it's",
    "forth",
    "只限",
    "被",
    "somewhere",
    "旁人",
    "设若",
    "ought",
    "结果",
    "有着",
    "before",
    "完全",
    "可见",
    "it'd",
    "愿意",
    "非常",
    "最高",
    "tell",
    "深入",
    "我的",
    "并",
    "hopefully",
    "她",
    "overall",
    "有些",
    "qv",
    "彻底",
    "以",
    "should",
    "那会儿",
    "从",
    "现在",
    "哪儿",
    "过",
    "later",
    "把",
    "where",
    "宁肯",
    "we've",
    "一直",
    "那边",
    "方面",
    "i'd",
    "嘻",
    "且",
    "behind",
    "having",
    "it",
    "by",
    "yourselves",
    "aren't",
    "有利",
    "did",
    "下来",
    "最後",
    "哎呀",
    "看看",
    "wouldn't",
    "彼",
    "third",
    "及其",
    "i've",
    "纵令",
    "自己",
    "呜",
    "替",
    "hereby",
    "焉",
    "显著",
    "较",
    "反之",
    "relatively",
    "附近",
    "few",
    "哎哟",
    "try",
    "thereafter",
    "这就是说",
    "best",
    "consequently",
    "contains",
    "好象",
    "they",
    "什麽",
    "凭借",
    "thus",
    "由",
    "伟大",
    "待",
    "kept",
    "saw",
    "seeming",
    "不论",
    "otherwise",
    "则",
    "却不",
    "那里",
    "战斗",
    "say",
    "much",
    "适当",
    "后面",
    "especially",
    "怎么样",
    "继而",
    "引起",
    "行动",
    "要",
    "嗳",
    "巨大",
    "他",
    "下列",
    "临",
    "任",
    "某些",
    "可",
    "冲",
    "那时",
    "obviously",
    "因此",
    "叫",
    "左右",
    "其次",
    "hence",
    "与否",
    "专门",
    "need",
    "巩固",
    "总而言之",
    "wish",
    "old",
    "前者",
    "做到",
    "往",
    "一些",
    "总是",
    "course",
    "以来",
    "something",
    "使用",
    "最大",
    "these",
    "主要",
    "among",
    "somebody",
    "where's",
    "本着",
    "at",
    "么",
    "new",
    "along",
    "不然",
    "俺们",
    "即使",
    "望",
    "thank",
    "明确",
    "ourselves",
    "否则",
    "共同",
    "her",
    "him",
    "等",
    "注意",
    "let",
    "way",
    "的",
    "倘使",
    "aside",
    "向着",
    "无论",
    "嘎",
    "要么",
    "目前",
    "有点",
    "怎样",
    "out",
    "逐渐",
    "throughout",
    "个别",
    "之类",
    "enough",
    "我",
    "综上所述",
    "yet",
    "倘若",
    "if",
    "随著",
    "both",
    "不久",
    "never",
    "另",
    "为什么",
    "hadn't",
    "wherein",
    "don't",
    "becomes",
    "此时",
    "比方",
    "靠",
    "既然",
    "anyway",
    "决定",
    "一次",
    "well",
    "各位",
    "既是",
    "anyone",
    "fifth",
    "吱",
    "相对",
    "促进",
    "whether",
    "主张",
    "确定",
    "所",
    "後来",
    "吧哒",
    "go",
    "什么",
    "sub",
    "怎麽",
    "起",
    "哪些",
    "例如",
    "presumably",
    "重新",
    "thanks",
    "喏",
    "限制",
    "安全",
    "咋",
    "here's",
    "强调",
    "indicates",
    "她们",
    "就是说",
    "through",
    "nevertheless",
    "only",
    "哇",
    "倘",
    "不只",
    "不同",
    "并没有",
    "要不",
    "呗",
    "you're",
    "尤其",
    "ones",
    "herself",
    "afterwards",
    "you",
    "各自",
    "有效",
    "moreover",
    "example",
    "得出",
    "之所以",
    "另外",
    "然而",
    "转变",
    "不变",
    "如上所述",
    "若",
    "shall",
    "哩",
    "如若",
    "沿着",
    "non",
    "顺",
    "其他",
    "nd",
    "及时",
    "allows",
    "不光",
    "有力",
    "everything",
    "goes",
    "直接",
    "哦",
    "倘或",
    "人家",
    "完成",
    "quite",
    "吗",
    "强烈",
    "hi",
    "does",
    "当着",
    "明显",
    "容易",
    "虽",
    "unto",
    "various",
    "地",
    "four",
    "no",
    "why",
    "总的来说",
    "经常",
    "trying",
    "whereafter",
    "况且",
    "take",
    "re",
    "从事",
    "能够",
    "该",
    "ltd",
    "been",
    "given",
    "be",
    "转动",
    "anything",
    "只是",
    "we",
    "together",
    "以及",
    "转贴",
    "那么些",
    "时候",
    "像",
    "哎",
    "心里",
    "为主",
    "不一",
    "或者",
    "above",
    "possible",
    "wonder",
    "further",
    "ok",
    "knows",
    "we'd",
    "既",
    "何处",
    "呀",
    "逐步",
    "still",
    "多少",
    "运用",
    "ever",
    "因为",
    "least",
    "onto",
    "didn't",
    "为何",
    "大力",
    "吓",
    "since",
    "因",
    "谁",
    "i'll",
    "今天",
    "follows",
    "尚且",
    "毋宁",
    "其余",
    "zz",
    "紧接着",
    "everybody",
    "再说",
    "作为",
    "自各儿",
    "许多",
    "这边",
    "那么",
    "一起",
    "不如",
    "正如",
    "everyone",
    "普遍",
    "come",
    "哈哈",
    "看来",
    "came",
    "containing",
    "本",
    "就",
    "regards",
    "不要",
    "不拘",
    "咦",
    "willing",
    "考虑",
    "不够",
    "互相",
    "咱们",
    "associated",
    "alone",
    "three",
    "according",
    "恰恰相反",
    "entirely",
    "喔唷",
    "咚",
    "打",
    "sure",
    "除非",
    "who",
    "那",
    "already",
    "following",
    "无宁",
    "look",
    "当",
    "应该",
    "非徒",
    "此外",
    "用",
    "perhaps",
    "每",
    "相似",
    "加之",
    "起来",
    "一时",
    "特别是",
    "changes",
    "你",
    "so",
    "about",
    "an",
    "indicated",
    "哉",
    "那个",
    "总的来看",
    "between",
    "upon",
    "已经",
    "another",
    "provides",
    "first",
    "they'd",
    "以上",
    "ain't",
    "这么",
    "几时",
    "self",
    "几",
    "either",
    "或",
    "经",
    "上来",
    "具体说来",
    "之前",
    "below",
    "greetings",
    "then",
    "without",
    "可以",
    "一致",
    "出去",
    "although",
    "our",
    "而已",
    "gets",
    "俺",
    "here",
    "around",
    "若是",
    "所有",
    "受到",
    "为了",
    "总结",
    "任凭",
    "这么点儿",
    "而是",
    "至于",
    "除了",
    "可能",
    "以为",
    "其",
    "多",
    "广泛",
    "downwards",
    "seven",
    "了",
    "嗯",
    "看出",
    "按",
    "able",
    "we're",
    "see",
    "until",
    "各个",
    "其它",
    "you'd",
    "即便",
    "矣",
    "双方",
    "rather",
    "novel",
    "看到",
    "beyond",
    "不可",
    "anybody",
    "want",
    "have",
    "got",
    "每个",
    "然後",
    "仍然",
    "无法",
    "”",
    "into",
    "趁",
    "如此",
    "such",
    "on",
    "实际",
    "认识",
    "有著",
    "whither",
    "呜呼",
    "seriously",
    "己",
    "别",
    "when",
    "them",
    "甚至",
    "brief",
    "应用",
    "接著",
    "突然",
    "looks",
    "same",
    "及至",
    "严格",
    "一般",
    "莫若",
    "whoever",
    "不是",
    "》",
    "即或",
    "cannot",
    "he's",
    "尽管",
    "真是",
    "之后",
    "also",
    "相对而言",
    "至",
    "down",
    "noone",
    "绝对",
    "其一",
    "was",
    "同",
    "得到",
    "please",
    "last",
    "comes",
    "however",
    "真正",
    "hither",
    "always",
    "有的",
    "the",
    "而况",
    "usually",
    "尽",
    "whereas",
    "nor",
    "先后",
    "的话",
    "just",
    "become",
    "th",
    "etc",
    "除此之外",
    "specifying",
    "of",
    "知道",
    "we'll",
    "shouldn't",
    "哪天",
    "等等",
    "立即",
    "you'll",
    "上去",
    "he",
    "know",
    "欢迎",
    "维持",
    "兮",
    "冒",
    "如其",
    "高兴",
    "from",
    "tried",
    "neither",
    "corresponding",
    "表明",
    "相信",
    "这麽",
    "密切",
    "不问",
    "看见",
    "进而",
    "一片",
    "上下",
    "达到",
    "上述",
    "开展",
    "can't",
    "正常",
    "为着",
    "腾",
    "于是乎",
    "sensible",
    "哼",
    "开始",
    "组成",
    "阿",
    "lately",
    "纵然",
    "或是",
    "appropriate",
    "whose",
    "今年",
    "呸",
    "six",
    "呵",
    "有",
    "far",
    "eight",
    "这时",
    "两者",
    "even",
    "yours",
    "合理",
    "然则",
    "抑或",
    "于",
    "whom",
    "became",
    "跟",
    "has",
    "thorough",
    "所谓",
    "once",
    "以至",
    "那儿",
    "取得",
    "在",
    "即令",
    "哪边",
    "清楚",
    "能否",
    "currently",
    "不但",
    "can",
    "那样",
    "过来",
    "other",
    "如下",
    "原来",
    "最好",
    "reasonably",
    "becoming",
    "yes",
    "over",
    "vs",
    "那些",
    "they'll",
    "namely",
    "比较",
    "其二",
    "welcome",
    "所以",
    "虽然",
    "叫做",
    "某",
    "联系",
    "anywhere",
    "更加",
    "何况",
    "彼此",
    "despite",
    "自身",
    "自个儿",
    "们",
    "嘿",
    "哈",
    "反映",
    "此间",
    "不能",
    "据",
    "之",
    "乌乎",
    "described",
    "将",
    "后来",
    "他人",
    "重大",
    "whereupon",
    "had",
    "谁知",
    "certainly",
    "各地",
    "其中",
    "哪样",
    "sent",
    "呢",
    "followed",
    "up",
    "问题",
    "多次",
    "全部",
    "在下",
    "nothing",
    "think",
    "遇到",
    "啦",
    "any",
    "宁愿",
    "通过",
    "whatever",
    "得",
    "行为",
    "and",
    "in",
    "有关",
    "they've",
    "while",
    "形成",
    "今後",
    "积极",
    "少数",
    "严重",
    "要不是",
    "这里",
    "herein",
    "连",
    "练习",
    "whence",
    "this",
    "叮咚",
    "管",
    "yourself",
    "成为",
    "这么些",
    "并不是",
    "没有",
    "接着",
    "many",
    "que",
    "tries",
    "seems",
    "过去",
    "万一",
    "十分",
    "第",
    "若非",
    "这样",
    "以後",
    "formerly",
    "基本",
    "普通",
    "但是",
    "everywhere",
    "t's",
    "何",
    "anyways",
    "什么样",
    "特殊",
    "为",
    "nowhere",
    "能",
    "were",
    "照",
    "不断",
    "具体",
    "适用",
    "避免",
    "不独",
    "嗬",
    "怎么办",
    "uses",
    "co",
    "oh",
    "怎么",
    "certain",
    "seemed",
    "其实",
    "采取",
    "也是",
    "across",
    "ZT",
    "get",
    "固然",
    "趁着",
    "啪达",
    "反应",
    "全面",
    "同时",
    "故",
    "thanx",
    "再者",
    "有所",
    "比",
    "虽说",
    "怎",
    "实现",
    "under",
    "specify",
    "何时",
    "到",
    "相反",
    "somewhat",
    "often",
    "how",
    "com",
    "tends",
    "those",
    "除",
    "former",
    "具体地说",
    "因而",
    "由此可见",
    "seen",
    "同一",
    "wants",
    "下面",
    "常常",
    "一面",
    "plus",
    "觉得",
    "具有",
    "wherever",
    "being",
    "instead",
    "并且",
    "un",
    "者",
    "must",
    "it'll",
    "获得",
    "边",
    "elsewhere",
    "沿",
    "也",
    "insofar",
    "accordingly",
    "better",
    "根本",
    "placed",
    "对于",
    "others",
    "这个",
    "necessary",
    "于是",
    "beforehand",
    "而且",
    "似乎",
    "unless",
    "以致",
    "上面",
    "加以",
    "代替",
    "mostly",
    "ignored",
    "不管",
    "thru",
    "within",
    "遭到",
    "自从",
    "不单",
    "whereby",
    "ie",
    "or",
    "unfortunately",
    "较之",
    "indeed",
    "selves",
    "不会",
    "而外",
    "设使",
    "构成",
    "正在",
    "而",
    "依靠",
    "eg",
    "来",
    "taken",
    "唉",
    "某个",
    "too",
    "－－",
    "hasn't",
    "自家",
    "during",
    "允许",
    "would",
    "probably",
    "seeing",
    "一旦",
    "to",
    "不足",
    "往往",
    "相同",
    "当时",
    "like",
    "gone",
    "按照",
    "是否",
    "secondly",
    "latterly",
    "zt",
    "里面",
    "这儿",
    "allow",
    "啥",
    "beside",
    "哗",
    "召开",
    "之一",
    "咱",
    "doesn't",
    "开外",
    "keeps",
    "nine",
    "越是",
    "someone",
    "最后",
    "满足",
    "纵",
    "ex",
    "乃至",
    "这",
    "weren't",
    "云云",
    "移动",
    "concerning",
    "前进",
    "总之",
    "又",
    "says",
    "going",
    "what",
    "ask",
    "朝着",
    "used",
    "需要",
    "存在",
    "its",
    "own",
    "争取",
    "大多数",
    "who's",
    "依照",
    "took",
    "朝",
    "来着",
    "more",
    "啊",
    "one",
    "好的",
    "以免",
    "只有",
    "以便",
    "said",
    "举行",
    "离",
    "all",
    "鄙人",
    "意思",
    "c'mon",
    "说明",
    "凭",
    "really",
    "ZZ",
    "merely",
    "thereupon",
    "available",
    "一下",
    "加强",
    "they're",
    "它",
    "每年",
    "though",
    "虽则",
    "向",
    "连同",
    "前后",
    "和",
    "sometimes",
    "同样",
    "causes",
    "name",
    "经过",
    "viz",
    "顺着",
    "上升",
    "done",
    "per",
    "然后",
    "对",
    "整个",
    "appreciate",
    "坚决",
    "mainly",
    "个",
    "theres",
    "方便",
    "less",
    "部分",
    "假如",
    "此",
    "相当",
    "itself",
    "是的",
    "what's",
    "使得",
    "以外",
    "there",
    "your",
    "必然",
    "还有",
    "他们",
    "喂",
    "inc",
    "赶",
    "cause",
    "变成",
    "是不是",
    "let's",
    "away",
    "near",
    "不特",
    "unlikely",
    "各种",
    "与其",
    "相应",
    "liked",
    "造成",
    "considering",
    "that's",
    "normally",
    "lest",
    "looking",
    "regarding",
    "鉴于",
    "particular",
    "认为",
    "howbeit",
    "second",
    "对应",
    "通常",
    "只要",
    "这么样",
    "maybe",
    "进行",
    "zero",
    "getting",
    "这种",
    "适应",
    "nearly",
    "坚持",
    "掌握",
    "haven't",
    "five",
    "甚而",
    "人们",
    "我们",
    "出现",
    "中间",
    "可是",
    "that",
    "which",
    "为什麽",
    "倘然",
    "after",
    "could",
    "认真",
    "着",
    "诸位",
    "结合",
    "是",
    "突出",
    "hers",
    "regardless",
    "以下",
    "处理",
    "but",
    "乘",
    "它们",
    "企图",
    "周围",
    "今后",
    "宁",
    "一定",
    "乃",
    "一切",
    "故此",
    "要是",
    "分别",
    "也罢",
    "using",
    "并不",
    "一则",
    "也好",
    "之後",
    "要不然",
    "rd",
    "哪年",
    "毫不",
    "任何",
    "couldn't",
    "hereafter",
    "总的说来",
    "那么样",
    "拿",
    "一来",
    "不得",
    "大约",
    "范围",
    "you've",
    "保持",
    "帮助",
    "off",
    "?",
    "will",
    "啐",
    "keep",
    "themselves",
    "哼唷",
    "重要",
    "whenever",
    "truly",
    "use",
    "mean",
    "一方面",
    "呕",
    "别说",
    "furthermore",
    "inward",
    "isn't",
    "说说",
    "very",
    "twice",
    "出来",
    "一边",
    "again",
    "即若",
    "不惟",
    "但",
    "从而",
    "she",
    "他的",
    "sometime",
    "相等",
    "任务",
    "以前",
    "极了",
    "呃",
    "his",
    "saying",
    "它们的",
    "不成",
    "准备",
    "--",
    "therein",
    "awfully",
    "thereby",
    "two",
}



import json
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Type, Union

from loguru import logger
from pydantic import BaseModel

from uglychain.llm import BaseLanguageModel
from uglychain.utils import config, retry_decorator

from .error import BadRequestError, RequestLimitError, Unauthorized


def not_notry_exception(exception: BaseException):
    if isinstance(exception, BadRequestError):
        return False
    if isinstance(exception, Unauthorized):
        return False
    return True


@dataclass
class ChatGLM(BaseLanguageModel):
    use_max_tokens: bool = False
    MAX_TOKENS: int = 128000
    top_p: float = field(init=False, default=0.7)
    # use_native_tools: bool = field(init=False, default=True)

    def generate(
        self,
        prompt: str = "",
        response_model: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Callable]] = None,
        stop: Union[Optional[str], List[str]] = None,
    ) -> str:
        kwargs = self.get_kwargs(prompt, response_model, tools, stop)
        if stop:
            if isinstance(stop, str):
                kwargs["stop"] = [stop]
            elif isinstance(stop, list):
                kwargs["stop"] = stop[0:1]
        else:
            kwargs.pop("stop")
        response = self.completion_with_backoff(**kwargs)
        logger.trace(f"kwargs:{kwargs}\nresponse:{response.choices[0].model_dump()}")
        if self.use_native_tools and response.choices[0].message.tool_calls:
            tool_calls_response = response.choices[0].message.tool_calls[0].function
            if tools:
                return json.dumps({"name": tool_calls_response.name, "args": json.loads(tool_calls_response.arguments)})
            elif response_model:
                return tool_calls_response.arguments
        return response.choices[0].message.content.strip()

    @property
    def default_params(self) -> Dict[str, Any]:
        kwargs = {
            "model": self.model,
            "do_sample": True,
            "temperature": self.temperature,
            "top_p": self.top_p,
        }
        if self.use_max_tokens:
            kwargs["max_tokens"] = self.max_tokens
        return kwargs

    def _create_client(self):
        try:
            from zhipuai import ZhipuAI
        except ImportError as err:
            raise ImportError("You need to install `pip install dashscope` to use this provider.") from err
        return ZhipuAI(api_key=config.zhipuai_api_key)

    @retry_decorator(not_notry_exception)
    def completion_with_backoff(self, **kwargs):
        from zhipuai import APIStatusError

        try:
            return self.client.chat.completions.create(**kwargs)
        except APIStatusError as error:
            status_code = error.status_code
            error_json = json.loads(error.response.text.strip())
            code = error_json.get("error").get("code")
            message = error_json.get("error").get("message")
            if status_code in [400, 404, 435]:
                # 400 Bad Request
                raise BadRequestError(f"code: {code}, message:{message}") from error
            elif status_code == 429 and code in ["1302", "1303", "1305"]:
                # 404 Not Found
                raise RequestLimitError(f"code: {code}, message:{message}") from error
            elif status_code in [401, 429, 434]:
                # 401 Unauthorized
                raise Unauthorized(f"code: {code}, message:{message}") from error
            else:
                raise error
        except Exception as e:
            raise e

    @property
    def max_tokens(self):
        tokens = self._num_tokens(messages=self.messages, model=self.model) + 1000  # add 1000 tokens for answers
        if not self.MAX_TOKENS > tokens:
            raise Exception(
                f"Prompt is too long. This model's maximum context length is {self.MAX_TOKENS} tokens. your messages required {tokens} tokens"
            )
        max_tokens = self.MAX_TOKENS - tokens + 1000
        return 2000 if max_tokens > 2000 else max_tokens



from typing import Any

from pydantic.v1 import Field, root_validator

try:
    from llama_index.core.llms import (
        CompletionResponse,
        CompletionResponseGen,
        CustomLLM,
        LLMMetadata,
    )
    from llama_index.core.llms.callbacks import llm_completion_callback
except ImportError as err:
    raise ImportError("Please install the `llama-index-core` package to use this LLM.") from err

from uglychain.provider import Model, get_llm_provider

from .base import BaseLanguageModel


class LlamaIndexLLM(CustomLLM):
    model: Model = Model.DEFAULT
    llm: BaseLanguageModel = Field(init=False, default=None)

    @root_validator(pre=False, skip_on_failure=True)
    def create_llm(cls, values):
        model_name = values.get("model")
        values["llm"] = get_llm_provider(model_name)
        return values

    @property
    def metadata(self) -> LLMMetadata:
        """Get LLM metadata."""
        if hasattr(self.llm, "MAX_TOKENS"):
            context_window = self.llm.MAX_TOKENS  # type: ignore
        else:
            context_window = 4096
        return LLMMetadata(
            context_window=context_window,
            is_chat_model=True,
            model_name=self.llm.model,
        )

    @llm_completion_callback()
    def complete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
        response = self.llm.generate(prompt)
        return CompletionResponse(text=response)

    @llm_completion_callback()
    def stream_complete(self, prompt: str, **kwargs: Any) -> CompletionResponseGen:
        response = ""
        dummy_response = self.llm.generate(prompt)
        for token in dummy_response:
            response += token
            yield CompletionResponse(text=response, delta=token)



from uglychain import LLM


def text_completion(prompt: str) -> str:
    """A tool that uses LLM to generate, summarize, and/or analyze text and code.

    Args:
        prompt (str): The prompt to use for text generation.
    """
    llm = LLM()
    return llm(prompt)



from .subprocess_language import SubprocessLanguage


class Python(SubprocessLanguage):
    file_extension = "py"
    name = "Python"

    def __init__(self):
        super().__init__()
        self.start_cmd = ["python3", "-iquB"]

    def preprocess_code(self, code):
        """
        Add active line markers
        Wrap in a tryCatch for better error handling in R
        Add end of execution marker
        """

        lines = code.split("\n")
        processed_lines = ['    print("##active_line1##")']

        for line in lines:
            processed_lines.append(f"    {line}")
        # Join lines to form the processed code
        processed_code = "\n".join(processed_lines)

        # Wrap in a tryCatch for error handling and add end of execution marker
        processed_code = f"""try:
{processed_code}
except Exception as e:
    print(f"##execution_error##{{e}}")

print("##end_of_execution##")
"""
        # Count the number of lines of processed_code
        # (R echoes all code back for some reason, but we can skip it if we track this!)
        self.code_line_count = len(processed_code.split("\n")) - 1

        return processed_code

    def detect_active_line(self, line):
        if "##active_line" in line:
            return int(line.split("##active_line")[1].split("##")[0])
        return None

    def detect_end_of_execution(self, line):
        return "##end_of_execution##" in line or "##execution_error##" in line



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any


@dataclass
class Storage(ABC):
    @abstractmethod
    def save(self, data: Any) -> None:
        pass

    @abstractmethod
    def load(self, *args, **kwargs) -> Any:
        pass



#!/usr/bin/env python3

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Type, Union

from loguru import logger
from pydantic import BaseModel

from uglychain.llm import BaseLanguageModel
from uglychain.utils import config, retry_decorator


@dataclass
class Gemini(BaseLanguageModel):
    use_max_tokens: bool = False

    def generate(
        self,
        prompt: str = "",
        response_model: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Callable]] = None,
        stop: Union[Optional[str], List[str]] = None,
    ) -> str:
        kwargs = self.get_kwargs(prompt, response_model, tools, stop)
        response = self.completion_with_backoff(**kwargs)
        logger.trace(f"kwargs:{kwargs}\nresponse:{response}")
        return response.choices[0].message.content.strip()

    # TODO: Set stop to stop_sequences in params
    def get_kwargs(
        self,
        prompt: str,
        response_model: Optional[Type],
        tools: Optional[List[Callable]],
        stop: Union[Optional[str], List[str]] = None,
    ) -> Dict[str, Any]:
        kwargs = super().get_kwargs(prompt, response_model, tools, stop)
        contents = kwargs["messages"]
        kwargs.pop("messages")
        kwargs["contents"] = contents
        return kwargs

    def _generate_messages(self, prompt: str) -> List[Dict[str, str]]:
        """Generate the list of messages for the conversation.

        Args:
            prompt (str): The user prompt.

        Returns:
            List[Dict[str, str]]: The list of messages.

        """
        if not self.is_continuous:
            self.messages = []
            if hasattr(self, "system_prompt") and self.system_prompt:
                self.messages.append({"role": "system", "parts": [self.system_prompt]})
        if prompt:
            self.messages.append({"role": "user", "parts": [prompt]})
        return self.messages

    @property
    def default_params(self) -> Dict[str, Any]:
        config = self.client.types.GenerationConfig(max_output_tokens=self.max_tokens, temperature=self.temperature)
        kwargs = {
            "generation_config": config,
        }
        return kwargs

    def _create_client(self):
        try:
            import google.generativeai as genai

            genai.configure(api_key=config.gemini_api_key, transport="rest")
        except ImportError as e:
            raise ImportError("You need to install `pip install google-generativeai` to use this provider.") from e
        return genai

    @retry_decorator()
    def completion_with_backoff(self, **kwargs):
        model = self.client.GenerativeModel(self.model)
        return model.generate_content(**kwargs)

    @property
    def max_tokens(self):
        return 2000



#!/usr/bin/env python3

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, Literal, Optional

from loguru import logger

from uglychain.storage import Storage

from .retrieve_with_llm import answer_with_llm, answer_with_map_llm, answer_with_reduce_llm

DEFAULT_N = 5


class BaseRetriever(ABC):
    default_n: int = DEFAULT_N

    @abstractmethod
    def search(self, query: str, n: int) -> List[str]:
        pass

    def get(
        self, query: str, response_mode: Literal["no_text", "compact", "refine", "summary"] = "no_text", n: int = 0
    ) -> str:
        if n == 0:
            n = self.default_n
        context = self.search(query, n)
        if response_mode != "no_text" and context:
            logger.trace(context)
        try:
            if response_mode == "no_text":
                return str(context)
            elif response_mode == "compact":
                return answer_with_llm(query, context)
            elif response_mode == "refine":
                return answer_with_reduce_llm(query, context)
            elif response_mode == "summary":
                return answer_with_map_llm(query, context)
        except Exception:
            return ""


@dataclass
class StorageRetriever(BaseRetriever, ABC):
    start_init: bool = False
    storage: Storage = field(init=False)

    def __post_init__(self):
        if self.start_init:
            self.init()
        else:
            self._load()

    @abstractmethod
    def init(self):
        pass

    @abstractmethod
    def add(self, text: str, metadata: Optional[Dict[str, str]] = None):
        pass

    def _load(self):
        self.init()

    def all(self) -> List[str]:
        return []



from typing import Any, Callable

from loguru import logger
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_random_exponential,
)


def retry_decorator(
    not_notry_exception: Callable[[BaseException], bool] = lambda x: False,
    max_retries: int = 6,
) -> Callable[[Any], Any]:
    min_seconds = 5
    max_seconds = 60
    return retry(
        reraise=True,
        stop=stop_after_attempt(max_retries),
        retry=retry_if_exception(not_notry_exception),
        wait=wait_random_exponential(min=min_seconds, max=max_seconds),
        before_sleep=before_sleep_log(logger, "WARNING"),  # type: ignore
    )



#!/usr/bin/env python3

import random
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union, cast

from loguru import logger
from pydantic import BaseModel

from .instructor import Instructor
from .tools import FUNCTION_CALL_FORMAT, FUNCTION_CALL_WITH_FINISH_FORMAT, FunctionCall, finish, tools_schema

TEMPERATURE = 0.3
FREQUENCY_PENALTY = 0
PRESENCE_PENALTY = 0
TOP_P = 1
T = TypeVar("T", bound=BaseModel)


@dataclass
class BaseLanguageModel(ABC):
    """Base class for LLM providers.

    This class is an abstract base class (ABC) for LLM providers. It defines the common interface that all LLM providers should implement.

    Attributes:
        delay_init (bool): Whether to delay the initialization of the LLM provider.
        model (str): The model used by the LLM provider.
        use_max_tokens (bool): Whether to use the maximum number of tokens when generating responses.
        client (Any): The client used by the LLM provider.
        temperature (float): The temperature parameter used for generating responses.
        system_prompt (Optional[str]): The system message to be displayed.
        is_continuous (bool): Whether the conversation is continuous.
        messages (list): The list of messages in the conversation.

    Methods:
        set_system_prompt: Set the system message.
        generate: Ask a question and return the user's response.
        completion_with_backoff: Perform completion with backoff.
        _default_params: Get the default parameters for generating responses.
        _create_client: Create the client for the LLM provider.
        _generate_validation: Perform validation before generating responses.
        _generate_messages: Generate the list of messages for the conversation.
        _num_tokens: Calculate the number of tokens in a conversation.
        max_tokens: Get the maximum number of tokens that can be returned at once.
    """

    is_init_delay: bool
    model: str
    client: Any = field(init=False)
    temperature: float = field(init=False, default=TEMPERATURE)
    top_p: float = field(init=False, default=TOP_P)
    frequency_penalty: float = field(init=False, default=FREQUENCY_PENALTY)
    presence_penalty: float = field(init=False, default=PRESENCE_PENALTY)
    system_prompt: Optional[str] = field(init=False, default=None)
    use_max_tokens: bool = field(init=False, default=False)
    seed: int = field(init=False, default=random.getrandbits(16))
    is_continuous: bool = field(init=False, default=False)
    messages: list = field(init=False, default_factory=list)
    use_native_tools: bool = field(init=False, default=False)

    def __post_init__(self):
        if not self.is_init_delay:
            self.client = self._create_client()

    def set_system_prompt(self, msg: Optional[str] = None) -> None:
        """Set the system message.

        Args:
            msg (str, optional): The message to set.

        Returns:
            None

        """
        if msg:
            self.system_prompt = msg

    @abstractmethod
    def generate(
        self,
        prompt: str = "",
        response_model: Optional[Type[T]] = None,
        tools: Optional[List[Callable]] = None,
        stop: Union[Optional[str], List[str]] = None,
    ) -> Any:
        """Ask a question and return the user's response.

        Args:
            prompt (str, optional): The question prompt.
            response_model (BaseModel, optional): The response model.

        Returns:
            str: The user's response.

        """
        kwargs = self.get_kwargs(prompt, response_model, tools, stop)
        response = self.completion_with_backoff(**kwargs)
        logger.trace(f"kwargs:{kwargs}\nresponse:{response}")
        return response

    def get_kwargs(
        self,
        prompt: str,
        response_model: Optional[Type[T]],
        tools: Optional[List[Callable]],
        stop: Union[Optional[str], List[str]],
    ) -> Dict[str, Any]:
        self._generate_validation()
        if tools:
            if self.use_native_tools:
                self._generate_messages(prompt)
                params = self.default_params
                params["tools"] = tools_schema(tools)
                if len(tools) == 1:
                    params["tool_choice"] = {"type": "function", "function": {"name": tools[0].__name__}}
                kwargs = {"messages": self.messages, "stop": stop, **params}
                return kwargs
            if not response_model:
                response_model = cast(Type[T], FunctionCall)
            try:
                index = tools.index(finish)
                tools[:] = [tools[index]] + tools[:index] + tools[index + 1 :]
                tool_names = ", ".join([f"`{tool.__name__}`" for tool in tools])
                prompt += f"\n{FUNCTION_CALL_WITH_FINISH_FORMAT.format(tool_names = tool_names, tool_schema = tools_schema(tools))}"
            except ValueError:
                tool_names = ", ".join([f"`{tool.__name__}`" for tool in tools])
                prompt += f"\n{FUNCTION_CALL_FORMAT.format(tool_names = tool_names, tool_schema = tools_schema(tools))}"
        if response_model:
            if self.use_native_tools:
                self._generate_messages(prompt)
                params = self.default_params
                schema = response_model.model_json_schema()
                params["tools"] = [
                    {
                        "type": "function",
                        "function": {
                            "name": schema["title"],
                            "description": f"Correctly extracted `{schema['title']}` with all the required parameters with correct types",
                            "parameters": {k: v for k, v in schema.items() if k not in ("title", "description")},
                        },
                    }
                ]
                params["tool_choice"] = {
                    "type": "function",
                    "function": {"name": schema["title"]},
                }
                kwargs = {"messages": self.messages, "stop": stop, **params}
                return kwargs
            instructor = Instructor.from_BaseModel(response_model)
            prompt += "\n" + instructor.get_format_instructions()
        self._generate_messages(prompt)
        if stop and isinstance(stop, str):
            stop = [stop]
        kwargs = {"messages": self.messages, "stop": stop, **self.default_params}
        return kwargs

    def parse_response(self, response: str, response_model: Type[T]) -> T:
        """Parse the response from the LLM provider.

        Args:
            response (str): The response from the LLM provider.
            response_model (BaseModel): The response model.

        Returns:
            BaseModel: The parsed response.

        """
        instructor = Instructor.from_BaseModel(response_model)
        return cast(T, instructor.from_response(response))

    @abstractmethod
    def completion_with_backoff(self, **kwargs):
        pass

    @property
    def default_params(self) -> Dict[str, Any]:
        """Get the default parameters for generating responses.

        Returns:
            Dict[str, Any]: The default parameters.

        """
        kwargs = {
            "model": self.model,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
            "seed": self.seed,
        }
        if self.use_max_tokens:
            kwargs["max_tokens"] = self.max_tokens
        return kwargs

    @abstractmethod
    def _create_client(self):
        """Create the client for the LLM provider.

        Returns:
            Any: The client.

        """
        pass

    def _generate_validation(self):
        """Perform validation before generating responses."""
        if self.is_init_delay:
            self.client = self._create_client()

    def _generate_messages(self, prompt: str) -> List[Dict[str, str]]:
        """Generate the list of messages for the conversation.

        Args:
            prompt (str): The user prompt.

        Returns:
            List[Dict[str, str]]: The list of messages.

        """
        if not self.is_continuous:
            self.messages = []
            if hasattr(self, "system_prompt") and self.system_prompt:
                self.messages.append({"role": "system", "content": self.system_prompt})
        if prompt:
            self.messages.append({"role": "user", "content": prompt})
        return self.messages

    def _num_tokens(self, messages: list, model: str):
        """Calculate the number of tokens in a conversation.

        Args:
            messages (list): The list of messages in the conversation.
            model (str): The model to use for tokenization.

        Returns:
            int: The number of tokens in the conversation.

        """
        return 0

    @property
    @abstractmethod
    def max_tokens(self) -> int:
        """Get the maximum number of tokens that can be returned at once.

        Returns:
            int: The maximum number of tokens.

        """
        pass



from .chains import LLM, MapChain, Model, ReActChain, ReduceChain
from .llm import BaseLanguageModel, finish, run_function
from .retrievers import BaseRetriever, Retriever, StorageRetriever
from .worker import BaseWorker

__version__ = "0.1.7"

__all__ = [
    "BaseLanguageModel",
    "Model",
    "finish",
    "run_function",
    "LLM",
    "MapChain",
    "ReduceChain",
    "ReActChain",
    "Retriever",
    "BaseRetriever",
    "StorageRetriever",
    "BaseWorker",
]



from .base import Storage
from .dill import DillStorage
from .llama_index import LlamaIndexStorage
from .sqlite import SQLiteStorage

__all__ = ["Storage", "SQLiteStorage", "DillStorage", "LlamaIndexStorage"]



from .base import Chain
from .llm import LLM, Model
from .map import MapChain
from .react_bad import ReActChain
from .reduce import ReduceChain

__all__ = [
    "Chain",
    "LLM",
    "Model",
    "MapChain",
    "ReduceChain",
    "ReActChain",
]



#!/usr/bin/env python3

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union

from loguru import logger
from pydantic import BaseModel

from uglychain.llm import BaseLanguageModel, ParseError
from uglychain.llm.tools import ActionResopnse, FunctionCall
from uglychain.provider import Model, get_llm_provider

from ..base import Chain
from .prompt import Prompt

GenericResponseType = TypeVar("GenericResponseType", bound=BaseModel)


@dataclass
class LLM(Chain, Generic[GenericResponseType]):
    """A class representing an LLM chain.

    This class inherits from the Chain class and implements the logic for interacting with a language model.

    Attributes:
        llm_name: The name of the language model.
        prompt_template: The template for the prompt.
    """

    prompt_template: str = "{prompt}"
    model: Model = Model.DEFAULT
    system_prompt: Optional[str] = None
    response_model: Optional[Type[GenericResponseType]] = None
    tools: Optional[List[Callable]] = None
    stop: Optional[Union[str, List[str]]] = None
    memory_callback: Optional[Callable[[Tuple[str, str]], None]] = None
    is_init_delay: bool = field(init=False, default=False)
    llm: BaseLanguageModel = field(init=False)

    def __post_init__(self):
        """Initialize the LLMChain object.

        This method initializes the LLMChain object by getting the LLM provider and setting the prompt.
        """
        # assert not (self.tools and self.response_model), "tools and response_model cannot be set at the same time"
        if self.tools:
            assert (
                self.response_model is None or self.response_model == ActionResopnse
            ), "response_model must be ActionResopnse if tools is set"
        self.llm = get_llm_provider(self.model, self.is_init_delay)
        logger.success(f"{self.model} loaded")
        if self.system_prompt:
            self.llm.set_system_prompt(self.system_prompt)
        self.prompt = self.prompt_template

    @property
    def input_keys(self) -> List[str]:
        """Get the input keys.

        Returns:
            A list of input keys.
        """
        return self.prompt.input_variables

    def _call(self, inputs: Dict[str, Any]) -> Union[GenericResponseType, str]:
        """Call the LLMChain.

        This method calls the LLMChain by formatting the prompt with the inputs and asking the LLM provider.

        Args:
            inputs: A dictionary of inputs.

        Returns:
            The response from the LLM provider.
        """
        prompt = self.prompt.format(**inputs)

        max_retries = 3  # 设置最大重试次数
        attempts = 0  # 初始化尝试次数
        while attempts < max_retries:
            try:
                response = self.llm.generate(prompt, self.response_model, self.tools, self.stop)
                if self.response_model:
                    instructor_response = self.llm.parse_response(response, self.response_model)
                    if self.tools and instructor_response.action.name not in [tool.__name__ for tool in self.tools]:  # type: ignore
                        raise ParseError(f"Tool {instructor_response.action.name} not found")  # type: ignore
                    if self.memory_callback:
                        self.memory_callback((prompt, response))
                    return instructor_response
                elif self.tools:
                    instructor_response = self.llm.parse_response(response, FunctionCall)
                    if instructor_response.name not in [tool.__name__ for tool in self.tools]:
                        raise ParseError(f"Tool {instructor_response.name} not found")
                    if self.memory_callback:
                        self.memory_callback((prompt, response))
                    return instructor_response  # type: ignore
                else:
                    if self.memory_callback:
                        self.memory_callback((prompt, response))
                    return response
            except ParseError as e:  # 捕获所有异常
                attempts += 1  # 尝试次数增加
                logger.warning("解析失败，正在尝试重新解析")
                logger.trace(f"第 {attempts} 次尝试解析失败，原因：{e}")
                if attempts == max_retries:
                    raise e  # 如果达到最大尝试次数，则抛出最后一个异常
            except Exception as e:
                raise e
        raise

    @property
    def prompt(self) -> Prompt:
        """Set the prompt.

        This method sets the prompt using the given prompt template.

        Args:
            prompt_template: The template for the prompt.
        """
        return self._prompt

    @prompt.setter
    def prompt(self, prompt_template: str) -> None:
        """Set the prompt.

        This method sets the prompt using the given prompt template.

        Args:
            prompt_template: The template for the prompt.
        """
        self._prompt = Prompt(prompt_template)

    @property
    def custom_model_name(self) -> str:
        """Get the custom model name."""
        return self.llm.model

    @custom_model_name.setter
    def custom_model_name(self, custom_model_name: str) -> None:
        """Set the custom model name."""
        self.llm.model = custom_model_name

    @property
    def temperature(self) -> float:
        """Get the temperature."""
        return self.llm.temperature

    @temperature.setter
    def temperature(self, temperature: float) -> None:
        """Set the temperature."""
        self.llm.temperature = temperature

    @property
    def top_p(self) -> float:
        """Get the top_p."""
        return self.llm.top_p

    @top_p.setter
    def top_p(self, top_p: float) -> None:
        """Set the top_p."""
        self.llm.top_p = top_p

    @property
    def frequency_penalty(self) -> float:
        """Get the frequency_penalty."""
        return self.llm.frequency_penalty

    @frequency_penalty.setter
    def frequency_penalty(self, frequency_penalty: float) -> None:
        """Set the frequency_penalty."""
        self.llm.frequency_penalty = frequency_penalty

    @property
    def presence_penalty(self) -> float:
        """Get the presence_penalty."""
        return self.llm.presence_penalty

    @presence_penalty.setter
    def presence_penalty(self, presence_penalty: float) -> None:
        """Set the presence_penalty."""
        self.llm.presence_penalty = presence_penalty

    @property
    def seed(self) -> int:
        """Get the seed."""
        return self.llm.seed

    @seed.setter
    def seed(self, seed: int) -> None:
        """Set the seed."""
        self.llm.seed = seed



#!/usr/bin/env python3

from dataclasses import dataclass, field
from typing import Any, Dict, List, Union

from loguru import logger
from pathos.multiprocessing import ProcessingPool as Pool

from uglychain.llm import ParseError

from .llm import LLM, FunctionCall, GenericResponseType


@dataclass
class MapChain(LLM[GenericResponseType]):
    prompt_template: str = "{map_key}"
    is_init_delay: bool = field(init=False, default=True)
    map_keys: List[str] = field(default_factory=lambda: ["map_key"])

    def _validate_inputs(self, inputs: Dict[str, Any]) -> None:
        self.num = len(inputs[self.map_keys[0]])
        for map_key in self.map_keys:
            self._validate_map_key(map_key, inputs)
        super()._validate_inputs(inputs)

    def _validate_map_key(self, map_key, inputs):
        assert (
            map_key in self.input_keys and isinstance(inputs[map_key], List) and len(inputs[map_key]) == self.num
        ), f"MapChain expects {map_key} to be a list of strings with the same length"

    def _call(self, inputs: Dict[str, Any]) -> List[Union[GenericResponseType, str]]:
        inputs_list = self._generate_inputs_list(inputs)

        with Pool() as pool:
            response = pool.map(self._map_func(inputs), inputs_list)

        results = self._process_results(response)

        return results

    def _generate_inputs_list(self, inputs):
        return [
            {
                **{map_key: inputs[map_key][i] for map_key in self.map_keys},
                **{"index": i},
            }
            for i in range(self.num)
        ]

    def _map_func(self, inputs):
        def func(input) -> Dict[str, Union[int, str, GenericResponseType]]:
            new_input: Dict = {k: v for k, v in inputs.items() if k not in self.map_keys}
            new_input.update({mapping_key: input[mapping_key] for mapping_key in self.map_keys})
            prompt = self.prompt.format(**new_input)
            max_retries = 3  # 设置最大重试次数
            attempts = 0  # 初始化尝试次数
            while attempts < max_retries:
                try:
                    response = self.llm.generate(prompt, self.response_model, self.tools, self.stop)
                    if self.response_model:
                        instructor_response = self.llm.parse_response(response, self.response_model).model_dump_json()
                        if self.memory_callback:
                            self.memory_callback((prompt, response))
                        logger.debug(f"MapChain: {input['index']} finished")
                        return {"index": input["index"], "result": instructor_response}
                    elif self.tools:
                        instructor_response = self.llm.parse_response(response, FunctionCall).model_dump_json()
                        if self.memory_callback:
                            self.memory_callback((prompt, response))
                        return {"index": input["index"], "result": instructor_response}  # type: ignore
                    else:
                        if self.memory_callback:
                            self.memory_callback((prompt, response))
                        logger.debug(f"MapChain: {input['index']} finished")
                        return {"index": input["index"], "result": response}
                except ParseError as e:  # 捕获所有异常
                    attempts += 1  # 尝试次数增加
                    logger.warning(f"解析失败，第 {attempts} 次尝试重新解析")
                    logger.trace(f"第 {attempts} 次尝试解析失败，原因：{e}")
                    if attempts == max_retries:
                        return {
                            "index": input["index"],
                            "result": "Error",
                        }  # 如果达到最大尝试次数，则抛出最后一个异常
                except Exception as e:
                    logger.warning(f"MapChain: {input['index']} failed with error: {e}")
                    return {"index": input["index"], "result": "Error"}
            return {"index": input["index"], "result": "Error"}

        return func

    def _process_results(self, results) -> List[Union[GenericResponseType, str]]:
        results = sorted(results, key=lambda x: x["index"])
        if self.response_model:
            new_results = []
            for result in results:
                if result["result"] == "Error":
                    new_results.append(result["result"])
                else:
                    new_results.append(self.response_model.model_validate_json(result["result"]))
            return new_results
        else:
            return [result["result"] for result in results]



from enum import Enum

from .arxiv import ArxivRetriever
from .base import BaseRetriever, StorageRetriever
from .bing import BingRetriever
from .bm25 import BM25Retriever
from .custom import CustomRetriever
from .llama_index import LlamaIndexRetriever, LlamaIndexStorageRetriever


class Retriever(Enum):
    Bing = BingRetriever
    Arxiv = ArxivRetriever
    LlamaIndex = LlamaIndexRetriever
    BM25 = BM25Retriever
    Combine = "CombineRetriever"
    Custom = CustomRetriever

    def __call__(self, *args, **kwargs) -> BaseRetriever:
        if self.name == "Combine":
            from .combine import CombineRetriever

            return CombineRetriever(*args, **kwargs)
        return self.value(*args, **kwargs)  # type: ignore

    def getStorage(self, *args, **kwargs) -> StorageRetriever:
        if self.name == "LlamaIndex":
            retriever = LlamaIndexStorageRetriever(*args, **kwargs)
        else:
            retriever = self(*args, **kwargs)
        if not isinstance(retriever, StorageRetriever):
            raise TypeError(f"{self.name} is not a StoresRetriever")
        return retriever



from typing import List

from uglychain.chains import LLM, MapChain, ReduceChain

REDUCE_PROMPT = """The original query is as follows: {query_str}
We have provided an existing answer: {history}
We have the opportunity to refine the existing answer (only if needed) with some more context below.
------------
{context_str}
------------
Given the new context, refine the original answer to better answer the query. If the context isn't useful, return the original answer.
Refined Answer:
"""

DEFAULT_PROMPT = """Context information is below.
---------------------
{context_str}
---------------------
Given the context information and not prior knowledge, answer the query.
Query: {query_str}
Answer:
"""

SUMMARY_PROMPT = """
Write a summary of the following. Try to use only the information provided. Try to include as many key details as possible.


{context_str}


SUMMARY:
"""

ROLE = """You are to answer the query based on the context. You need to answer with the same language as the question use."""


def answer_with_llm(
    query: str,
    context: List[str],
):
    llm = LLM(DEFAULT_PROMPT, system_prompt=ROLE)
    return llm(query_str=query, context_str=context)


def answer_with_reduce_llm(
    query: str,
    context: List[str],
):
    llm = ReduceChain(REDUCE_PROMPT, reduce_keys=["context_str"], system_prompt=ROLE)
    return llm(query_str=query, context_str=context)


def answer_with_map_llm(
    query: str,
    context: List[str],
):
    llm = MapChain(SUMMARY_PROMPT, map_keys=["context_str"])
    summaries = llm(context_str=context)
    return answer_with_llm(query, summaries)
